constant euwhat_ebarcode_ex = "@(!--#) @(#) ebarcode.ex, version 023, 02-april-2002"

--
--  ebarcode.ex
--
--  create barcode labels as monochrome bitmaps
--

-----------------------------------------------------------------------------

--
--  includes
--

include wildcard.e

--
--  constants
--

include cmap.e

constant MAXLABEL = 8

constant BCHEIGHT = 64

constant BCWIDTHS = { 96 , 128 , 160 , 192 , 192 , 224 , 256 , 288 }

constant BCMARGINS = { 10 , 13 , 16 , 19 , 6 , 9 , 12 , 15 }

constant LEADER  = "100101101101"
constant TRAILER = "0100101101101"

constant BC = {
                "0-0101001101101" ,
                "1-0110100101011" ,
                "2-0101100101011" ,
                "3-0110110010101" ,
                "4-0101001101011" ,
                "5-0110100110101" ,
                "6-0101100110101" ,
                "7-0101001011011" ,
                "8-0110100101101" ,
                "9-0101100101101" ,
                "A-0110101001011" ,
                "B-0101101001011" ,
                "C-0110110100101" ,
                "D-0101011001011" ,
                "E-0110101100101" ,
                "F-0101101100101" ,
                "G-0101010011011" ,
                "H-0110101001101" ,
                "I-0101101001101" ,
                "J-0101011001101" ,
                "K-0110101010011" ,
                "L-0101101010011" ,
                "M-0110110101001" ,
                "N-0101011010011" ,
                "O-0110101101001" ,
                "P-0101101101001" ,
                "Q-0101010110011" ,
                "R-0110101011001" ,
                "S-0101101011001" ,
                "T-0101011011001" ,
                "U-0110010101011" ,
                "V-0100110101011" ,
                "W-0110011010101" ,
                "X-0100101101011" ,
                "Y-0110010110101" ,
                "Z-0100110110101"
              }


--
--  globals
--

sequence progname

-----------------------------------------------------------------------------

procedure putbyte(integer file, integer byte)

  sequence buf

  buf = repeat(0, 1)

  buf[1] = byte

  puts(file, buf)

  return

end procedure

-----------------------------------------------------------------------------

procedure putword(integer file, integer word)

  putbyte(file, and_bits(word, #FF))

  putbyte(file, and_bits(word, #FF00) / #100)

  return

end procedure

-----------------------------------------------------------------------------

procedure putdword(integer file, integer dword)

  putbyte(file, and_bits(dword, #FF))

  putbyte(file, and_bits(dword, #FF00) / #100)

  putbyte(file, and_bits(dword, #FF0000) / #10000)

  putbyte(file, and_bits(dword, #FF000000) / #1000000)

  return

end procedure  

-----------------------------------------------------------------------------

procedure writebitmap(sequence bitmap, sequence filename)

  integer file
  integer h
  integer w
  integer byte

  h = length(bitmap)

  if remainder(h, 8) != 0 then
    printf(2, "%s: height of bitmap (%d) is not a multiple of 8\n", {progname, h})
    abort(2)
  end if

  w = length(bitmap[1])

  if remainder(w, 8) != 0 then
    printf(2, "%s: width of bitmap (%d) is not a multiple of 8\n", {progname, w})
    abort(2)
  end if

  file = open(filename, "wb")

  if file = -1 then
    printf(2, "%s: cannot open \"%s\" for binary write\n", {progname, filename})
    abort(2)
  end if

  -- offset 0-1          "BM" magic number
  putbyte(file, 'B')
  putbyte(file, 'M')

  -- offset 2-5          size of entire file in bytes
  putdword(file, 62 + (h * (w / 8)))

  -- offset 6-9          reserved for future use
  putdword(file, 0)

  -- offset 10-13        offset to bitmap data field
  putdword(file, 62)

  -- offset 14-17        length of bitmap header field
  putdword(file, 40)

  -- offset 18-21        width of bitmap in pixels
  putdword(file, w)

  -- offset 22-25        height of bitmap in pixels
  putdword(file, h)

  -- offset 26-27        number of planes in bitmap
  putword(file, 1)

  -- offset 28-29        bitmap type (1=monochrome)
  putword(file, 1)

  -- offset 30-33        compressions specifications (0=no compression)
  putdword(file, 0)

  -- offset 34-37        bitmap data size
  putdword(file, (h * (w / 8)))

  -- offset 38-41        horizontal resolution
  putdword(file, 0)

  -- offset 42-45        vertical resolution
  putdword(file, 0)

  -- offset 46-49        number of used colours (2=monochrome)
  putdword(file, 2)

  -- offset 50-53        number of important colours (2=monochrome)
  putdword(file, 2)

  -- offset 54-57        pallette spec (first colour)
  putbyte(file, #FF)
  putbyte(file, #FF)
  putbyte(file, #FF)
  putbyte(file, #00)

  -- offset 58-61        pallette spec (second colour)
  putbyte(file, #00)
  putbyte(file, #00)
  putbyte(file, #00)
  putbyte(file, #00)


  for x = length(bitmap) to 1 by -1 do
    for y = 1 to length(bitmap[x]) by 8 do
      byte = 0

      for b = 0 to 7 do
        byte = byte * 2
        byte = byte + bitmap[x][y+b]
--        printf(1, "%d", {bitmap[x][y+b]})
      end for

      putbyte(file, byte)
    end for
--    printf(1, "\n", {})
  end for      

  close(file)
  
end procedure

-----------------------------------------------------------------------------

function putline(sequence bitmap, sequence line, integer lft, integer cfl, integer step)

  integer val

  for i = 1 to length(line) do
    if line[i] = '1' then
      val = 1
    else
      val = 0
    end if

    for s = 1 to step do
      bitmap[lft][cfl+((i-1)*step)+(s-1)] = val
    end for
  end for

  return(bitmap)

end function

-----------------------------------------------------------------------------

function overlay(sequence mainbm, sequence smallbm, integer lft, integer cfl)

  for line = 1 to length(smallbm) do
    for col = 1 to length(smallbm[1]) do
      mainbm[lft+(line-1)][cfl+(col-1)] = smallbm[line][col]
    end for
  end for

  return(mainbm)

end function

-----------------------------------------------------------------------------

function lookupbc(sequence bc, integer c)

  sequence matched

  c = upper(c)

  matched = {}

  for i = 1 to length(bc) do
    if c = bc[i][1] then
      matched = bc[i][3..15]
    end if
  end for

  if equal(matched, {}) then
    printf(2, "%s: could not look up barcode for character \"%s\"\n", {progname, c})
    abort(2)
  end if

  return(matched)

end function

-----------------------------------------------------------------------------

function lookupcmap(sequence cmap, integer c)

  sequence matched

  c = upper(c)

  matched = {}

  for i = 1 to length(cmap) do
    if c = cmap[i][1] then
      matched = cmap[i][2]
    end if
  end for

  if equal(matched, {}) then
    printf(2, "%s: could not lookup character map for character \"%s\"\n", {progname, c})
    abort(2)
  end if

  return(matched)

end function

-----------------------------------------------------------------------------

procedure ebarcode(sequence label)

  sequence bitmap
  sequence filename
  sequence bc
  sequence cmap

  if length(label) > MAXLABEL then
    printf(2, "%s: label \"%s\" is too long (exceeds %d characters)\n", {progname, label, MAXLABEL})
    abort(2)
  end if

  bitmap = {}

  for i = 1 to BCHEIGHT do
    bitmap = append(bitmap, repeat(0, BCWIDTHS[length(label)]))
  end for

  bc = ""

  bc = bc & LEADER

  for i = 1 to length(label) do
    bc = bc & lookupbc(BC, label[i])
  end for

  bc = bc & TRAILER

  for i = 2 to (BCHEIGHT-11) do
    bitmap = putline(bitmap, bc, i, (BCMARGINS[length(label)])+1, 2)
  end for

  for i = 1 to length(label) do
    cmap = lookupcmap(CMAP, label[i])

    bitmap = overlay(bitmap, cmap, BCHEIGHT-9, (BCMARGINS[length(label)])+32+((i-1)*26))
  end for

  filename = label & ".bmp"

  writebitmap(bitmap, filename)

  return

end procedure

-----------------------------------------------------------------------------

--
--  Main
--

function main()

  integer  argc
  sequence argv

  argv = command_line()
  argc = length(argv)

  progname = "ebarcode"

  if (argc < 3) then
    printf(2, "%s: no barcode text specified on command line\n", {progname})
    abort(2)
  end if

  if (argc = 3) then
    if equal(argv[3], "--version") then
      printf(2, "\n", {})
      printf(2, "%s\n", {euwhat_ebarcode_ex})
      printf(2, "\n", {})
      printf(2, "    It may be warriors who get the glory\n", {})
      printf(2, "\n", {})
      printf(2, "    But engineers build the future\n", {})
      printf(2, "\n", {})

      abort(2)
    end if
  end if

  for arg = 3 to argc do
    ebarcode(argv[arg])
  end for

  return(0)

end function

-----------------------------------------------------------------------------

--
--  top level stub
--

abort(main())

-- end of file --
